'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _react = require('react');

var _react2 = _interopRequireDefault(_react);

var _propTypes = require('prop-types');

var _propTypes2 = _interopRequireDefault(_propTypes);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { 'default': obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

var StickyTree = function (_React$PureComponent) {
    _inherits(StickyTree, _React$PureComponent);

    function StickyTree(props) {
        _classCallCheck(this, StickyTree);

        var _this = _possibleConstructorReturn(this, (StickyTree.__proto__ || Object.getPrototypeOf(StickyTree)).call(this, props));

        _this.onScroll = _this.onScroll.bind(_this);

        _this.state = {
            currNodePos: 0
        };

        _this.scrollTop = 0;
        _this.nodePosCache = [];
        _this.rowRenderRange = undefined;
        return _this;
    }

    /**
     *  Converts the consumer's tree structure into a flat array with root at index: 0,
     *  including information about the top and height of each node.
     *
     *  i.e:
     *  [
     *    { id: 'root', top: 0, index: 0, height: 100 },
     *    { id: 'child1', top: 10, index: 0, parentIndex: 0 height: 10 },
     *    ...
     *  ]
     */


    _createClass(StickyTree, [{
        key: 'flattenTree',
        value: function () {
            function flattenTree(node) {
                var nodes = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
                var context = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : { totalHeight: 0, parentIndex: undefined };

                var index = nodes.length;
                var height = node.height;

                var nodeInfo = {
                    id: node.id,
                    isSticky: node.isSticky || false,
                    stickyTop: node.stickyTop || 0,
                    zIndex: node.zIndex || 0,
                    top: context.totalHeight,
                    parentIndex: context.parentIndex,
                    height: height,
                    index: index
                };

                nodes.push(nodeInfo);

                if (context.parentIndex !== undefined) {
                    nodes[context.parentIndex].children.push(index);
                }

                context.totalHeight += height;

                var children = this.props.getChildren(node.id);
                if (Array.isArray(children)) {
                    nodeInfo.children = [];
                    for (var i = 0; i < children.length; i++) {
                        // Need to reset parentIndex here as we are recursive.
                        context.parentIndex = index;
                        var child = children[i];
                        this.flattenTree(child, nodes, context);
                    }
                }

                nodeInfo.totalHeight = context.totalHeight - nodeInfo.top;

                return nodes;
            }

            return flattenTree;
        }()
    }, {
        key: 'componentWillMount',
        value: function () {
            function componentWillMount() {
                this.recomputeTree();
            }

            return componentWillMount;
        }()
    }, {
        key: 'componentWillReceiveProps',
        value: function () {
            function componentWillReceiveProps(newProps) {
                if (newProps.root !== this.props.root) {
                    this.nodePosCache = this.flattenTree(newProps.root);
                }

                if (newProps.scrollTop !== undefined && newProps.scrollTop >= 0 && newProps.scrollTop !== this.scrollTop) {
                    this.elem.scrollTop = newProps.scrollTop;
                }

                if (newProps.scrollIndex !== undefined && newProps.scrollIndex >= 0) {
                    this.scrollIndexIntoView(newProps.scrollIndex);
                }
            }

            return componentWillReceiveProps;
        }()
    }, {
        key: 'getNodeIndex',
        value: function () {
            function getNodeIndex(id) {
                // TODO: Might be best to create a lookup to support this.
                for (var i = 0, l = this.nodePosCache.length; i < l; ++i) {
                    if (this.nodePosCache[i].id === id) {
                        return i;
                    }
                }
                return -1;
            }

            return getNodeIndex;
        }()
    }, {
        key: 'isNodeInView',
        value: function () {
            function isNodeInView(node) {
                return this.isIndexInView(this.getNodeIndex(node));
            }

            return isNodeInView;
        }()

        /**
         * Returns true if the node is within the view. Note this this will return FALSE for visible sticky nodes that are partially out of
         * view disregarding sticky, which is useful when the node will become unstuck. This may occur when the node is collapsed in a tree.
         * In this case, you want to scroll this node back into view so that the collapsed node stays in the same position.
         *
         * @param index
         * @returns {boolean}
         */

    }, {
        key: 'isIndexInView',
        value: function () {
            function isIndexInView(index) {
                var node = this.nodePosCache[index];
                return this.elem.scrollTop <= node.top - node.stickyTop;
            }

            return isIndexInView;
        }()
    }, {
        key: 'scrollNodeIntoView',
        value: function () {
            function scrollNodeIntoView(node) {
                this.scrollIndexIntoView(this.getNodeIndex(node));
            }

            return scrollNodeIntoView;
        }()
    }, {
        key: 'scrollIndexIntoView',
        value: function () {
            function scrollIndexIntoView(scrollIndex) {
                var node = this.nodePosCache[scrollIndex];
                if (node !== undefined) {
                    this.elem.scrollTop = node.top - node.stickyTop;
                }
            }

            return scrollIndexIntoView;
        }()
    }, {
        key: 'componentDidUpdate',
        value: function () {
            function componentDidUpdate(prevProps, prevState) {
                if (this.props.onRowsRendered !== undefined && prevState.currNodePos !== this.props.currNodePos) {
                    var range = this.rowRenderRange;
                    var visibleStartInfo = this.nodePosCache[range.visibleStart];
                    var visibleEndInfo = this.nodePosCache[range.visibleEnd];

                    this.props.onRowsRendered({
                        overscanStartIndex: range.start,
                        overscanStopIndex: range.end,
                        startIndex: range.visibleStart,
                        stopIndex: range.visibleEnd,
                        startNode: visibleStartInfo && visibleStartInfo.id,
                        endNode: visibleEndInfo && visibleEndInfo.id
                    });
                }
            }

            return componentDidUpdate;
        }()
    }, {
        key: 'recomputeTree',
        value: function () {
            function recomputeTree() {
                if (this.props.root !== undefined) {
                    this.nodePosCache = this.flattenTree(this.props.root);
                    // Need to re-render as the curr node may not be in view
                    if (this.elem) {
                        // We need to find the the closest node to where we are scrolled to since the structure of the
                        // the tree probably has changed.
                        this.findClosestNode(this.elem.scrollTop, 0);
                    }
                    this.forceUpdate();
                }
            }

            return recomputeTree;
        }()
    }, {
        key: 'getChildContainerStyle',
        value: function () {
            function getChildContainerStyle(child, top) {
                return { position: 'absolute', top: top, height: child.totalHeight, width: '100%' };
            }

            return getChildContainerStyle;
        }()
    }, {
        key: 'getClientNodeStyle',
        value: function () {
            function getClientNodeStyle(node) {
                var style = { height: node.height };
                if (node.isSticky) {
                    style.position = 'sticky';
                    style.top = node.stickyTop;
                    style.zIndex = node.zIndex;
                }
                return style;
            }

            return getClientNodeStyle;
        }()
    }, {
        key: 'renderParentTree',
        value: function () {
            function renderParentTree() {
                this.rowRenderRange = this.getRenderRowRange();
                var path = this.getParentPath(this.rowRenderRange.start);

                // Parent nodes to the current range.
                var indexesToRender = new Set();
                for (var i = 0; i < path.length; i++) {
                    indexesToRender.add(path[i].index);
                }

                // The rest of the nodes within the range.
                for (var _i = this.rowRenderRange.start; _i <= this.rowRenderRange.end; _i++) {
                    indexesToRender.add(this.nodePosCache[_i].index);
                }

                if (this.props.renderRoot) {
                    return _react2['default'].createElement(
                        'div',
                        { className: 'rv-sticky-node-list', style: { width: '100%', position: 'absolute', top: 0 } },
                        this.renderChildWithChildren(path[0], 0, indexesToRender)
                    );
                }
                return this.renderParentContainer(path[0], indexesToRender);
            }

            return renderParentTree;
        }()
    }, {
        key: 'renderParentContainer',
        value: function () {
            function renderParentContainer(parent, indexesToRender) {
                return _react2['default'].createElement(
                    'div',
                    { key: parent.id, className: 'rv-sticky-node-list', style: { position: 'absolute', width: '100%' } },
                    this.renderChildren(parent, indexesToRender)
                );
            }

            return renderParentContainer;
        }()
    }, {
        key: 'renderChildWithChildren',
        value: function () {
            function renderChildWithChildren(child, top, indexesToRender) {
                return _react2['default'].createElement(
                    'div',
                    { className: 'rv-sticky-parent-node', key: child.id, style: this.getChildContainerStyle(child, top) },
                    this.props.rowRenderer({ id: child.id, style: this.getClientNodeStyle(child) }),
                    this.renderParentContainer(child, indexesToRender)
                );
            }

            return renderChildWithChildren;
        }()
    }, {
        key: 'renderChildren',
        value: function () {
            function renderChildren(parent, indexesToRender) {
                var _this2 = this;

                var nodes = [];
                var top = 0;
                parent.children.forEach(function (index) {
                    var child = _this2.nodePosCache[index];

                    if (indexesToRender.has(index)) {
                        if (child.children) {
                            nodes.push(_this2.renderChildWithChildren(child, top, indexesToRender));
                        } else {
                            nodes.push(_react2['default'].createElement(
                                'div',
                                {
                                    className: 'rv-sticky-leaf-node', key: child.id,
                                    style: _this2.getChildContainerStyle(child, top) },
                                _this2.props.rowRenderer({ id: child.id, style: _this2.getClientNodeStyle(child) })
                            ));
                        }
                    }
                    // Needs to be on the outside so that we add the the top even if
                    // this node is not visible
                    top += child.totalHeight;
                });
                return nodes;
            }

            return renderChildren;
        }()

        /**
         * Determines the start and end number of the range to be rendered.
         * @returns {{start: number, end: number}} Indexes within nodePosCache
         */

    }, {
        key: 'getRenderRowRange',
        value: function () {
            function getRenderRowRange() {
                var start = this.state.currNodePos - this.props.overscanRowCount;
                if (start < 0) {
                    start = 0;
                }
                var visibleEnd = this.state.currNodePos + 1;

                while (this.nodePosCache[visibleEnd] && this.nodePosCache[visibleEnd].top < this.scrollTop + this.props.height) {
                    visibleEnd++;
                }

                var end = visibleEnd + this.props.overscanRowCount;
                if (end > this.nodePosCache.length - 1) {
                    end = this.nodePosCache.length - 1;
                }

                return { start: start, end: end, visibleStart: this.state.currNodePos, visibleEnd: visibleEnd };
            }

            return getRenderRowRange;
        }()

        /**
         * Returns the parent path from nodePosCache for the specified index within nodePosCache.
         * @param nodeIndex
         * @returns {Array<Node>}
         */

    }, {
        key: 'getParentPath',
        value: function () {
            function getParentPath(nodeIndex) {
                var currNode = this.nodePosCache[nodeIndex];
                var path = [currNode];
                while (currNode) {
                    currNode = this.nodePosCache[currNode.parentIndex];
                    if (currNode) {
                        path.push(currNode);
                    }
                }
                return path.reverse();
            }

            return getParentPath;
        }()

        /**
         * Searches from the current node position downwards to see if the top of nodes above are greater
         * than or equal to the current scrollTop
         * @param scrollTop
         * @param searchPos
         * @returns {number}
         */

    }, {
        key: 'forwardSearch',
        value: function () {
            function forwardSearch(scrollTop, searchPos) {
                var nodePosCache = this.nodePosCache;
                for (var i = searchPos; i < nodePosCache.length; i++) {
                    if (nodePosCache[i].top >= scrollTop) {
                        return i;
                    }
                }
                return nodePosCache.length - 1;
            }

            return forwardSearch;
        }()

        /**
         * Searches from the current node position upwards to see if the top of nodes above are less than
         * or equal the current scrollTop.
         * @param scrollTop
         * @param searchPos
         * @returns {number}
         */

    }, {
        key: 'backwardSearch',
        value: function () {
            function backwardSearch(scrollTop, searchPos) {
                var nodePosCache = this.nodePosCache;
                for (var i = searchPos; i >= 0; i--) {
                    if (nodePosCache[i].top <= scrollTop) {
                        return i;
                    }
                }
                return 0;
            }

            return backwardSearch;
        }()

        /**
         * Returns the closest node within nodePosCache.
         * @param scrollTop
         * @param currNodePos
         */

    }, {
        key: 'findClosestNode',
        value: function () {
            function findClosestNode(scrollTop, currNodePos) {
                var pos = void 0;
                if (scrollTop > this.scrollTop || currNodePos === 0) {
                    pos = this.forwardSearch(scrollTop, currNodePos);
                }
                if (scrollTop < this.scrollTop || pos === undefined) {
                    pos = this.backwardSearch(scrollTop, currNodePos);
                }
                this.setState({ currNodePos: pos });
            }

            return findClosestNode;
        }()
    }, {
        key: 'onScroll',
        value: function () {
            function onScroll(e) {
                var scrollTop = e.target.scrollTop;
                this.findClosestNode(scrollTop, this.state.currNodePos);
                this.scrollTop = scrollTop;

                if (this.props.onScroll !== undefined) {
                    this.props.onScroll({ scrollTop: this.scrollTop });
                }
            }

            return onScroll;
        }()
    }, {
        key: 'render',
        value: function () {
            function render() {
                var _this3 = this;

                var style = { overflow: 'auto', position: 'relative' };
                if (this.props.width) {
                    style.width = this.props.width;
                }
                if (this.props.height) {
                    style.height = this.props.height;
                }

                return _react2['default'].createElement(
                    'div',
                    { ref: function () {
                            function ref(elem) {
                                return _this3.elem = elem;
                            }

                            return ref;
                        }(), className: 'rv-sticky-tree', style: style, onScroll: this.onScroll },
                    this.renderParentTree()
                );
            }

            return render;
        }()
    }]);

    return StickyTree;
}(_react2['default'].PureComponent);

StickyTree.propTypes = {
    getChildren: _propTypes2['default'].func.isRequired,

    /**
     * Called to retrieve a row to render. The function should return a single React node.
     * The function is called with an object in the form:
     *
     * <pre>
     *     rowRenderer({ id, style })
     * </pre>
     *
     * The id is the id from either the root property passed to the tree, or one returned in the getChildren call.
     */
    rowRenderer: _propTypes2['default'].func.isRequired,

    /**
     * An object which represents the root node in the form:
     *
     * {
     *   id: 'myRootNodeId',
     *   isSticky: true
     * }
     *
     * This will be the first node passed to rowRenderer({id: myRootNodeId, ...}). Your id might be an index in an array or map lookup.
     */
    root: _propTypes2['default'].object.isRequired,
    overscanRowCount: _propTypes2['default'].number.isRequired,
    height: _propTypes2['default'].number,
    width: _propTypes2['default'].number,
    renderRoot: _propTypes2['default'].bool,

    /**
     * Sets the position of the tree to the specified scrollTop. To reset
     * this, change this to -1 or undefined
     */
    scrollTop: _propTypes2['default'].number,

    /**
     * Sets the position of the tree to the specified scrollIndex. This is useful when
     * paired with onRowsRendered() which returns the startIndex and stopIndex.
     */
    scrollIndex: _propTypes2['default'].number,

    /**
     * Called whenever the scrollTop position changes.
     */
    onScroll: _propTypes2['default'].func,

    /**
     * Called to indicate that a new render range for rows has been rendered.
     */
    onRowsRendered: _propTypes2['default'].func
};
StickyTree.defaultProps = {
    overscanRowCount: 0,
    renderRoot: true
};
exports['default'] = StickyTree;